/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"
#include "apr_lib.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_provider.h"

#include "mod_smtpd.h"
#include "mod_smtpd_auth.h"

module AP_MODULE_DECLARE_DATA smtpd_auth_module;

typedef struct {
    int enable;
    apr_hash_t *methods;
} smtpd_auth_config;

static void str_toupper(char *str)
{
    while (*str) {
	*str = apr_toupper(*str);
	++str;
    }
}

static void *smtpd_auth_config_new(apr_pool_t *p,
				   server_rec *s)
{
    smtpd_auth_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_auth_config));

    cfg->enable = -1;
    cfg->methods = NULL;

    return cfg;
}

static void *smtpd_auth_config_merge(apr_pool_t *p, void *b,
				     void *n)
{
    smtpd_auth_config *base = (smtpd_auth_config *)b;
    smtpd_auth_config *new = (smtpd_auth_config *)n;

    smtpd_auth_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_auth_config));

    if (new->enable == -1) {
	cfg->enable = base->enable;
    }
    else {
	cfg->enable = new->enable;
    }

    cfg->methods = apr_hash_overlay(p, base->methods, new->methods);

    return cfg;
}

static const char *smtpd_auth_enable(cmd_parms *cmd,
				     void *in_dconf, int flag)
{
    smtpd_auth_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_auth_module);

    conf->enable = flag;

    return NULL;
}

static const command_rec smtpd_auth_cmds[] =
{
    AP_INIT_FLAG("SmtpAuthEnable", smtpd_auth_enable,
		 (void *) NULL, RSRC_CONF,
		 "The name of the auth modules to enable"),
    {NULL}
};

smtpd_retcode smtpd_auth_unrecognized_command(smtpd_conn_rec *scr, smtpd_return_data *in,
					      char *command, char *data)
{
    smtpd_auth_config *conf = ap_get_module_config(scr->s->module_config, &smtpd_auth_module);
    smtpd_retcode ret = SMTPD_DECLINED;

    if (conf != NULL && strcmp(command, "auth") == 0) {
	apr_pool_t *pool;
	unsigned char *method;
	char *buffer = data;

	smtpd_auth_method_st *handle_func;

	apr_pool_create(&pool, scr->p);

	method = ap_getword_white_nc(pool, &buffer);
	method = apr_pstrdup(pool, method);
	str_toupper(method);

	handle_func = apr_hash_get(conf->methods, method, APR_HASH_KEY_STRING);

	if (handle_func != NULL) {
	    int method_ret = handle_func->func(scr, command, in, buffer);
	    ret = SMTPD_DONE;

	    if (method_ret == SMTPD_DECLINED) {
		smtpd_respond_oneline(scr, 503, "AUTH not available");

		ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			     "[mod_smtpd_auth.c] smtpd_auth_unrecognized_command: auth %s not available!", method);

	    }
	    else if (in->msgs == NULL) {
		smtpd_respond_oneline(scr, 503, "AUTH not available");

		ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			     "[mod_smtpd_auth.c] smtpd_auth_unrecognized_command: handler did not set a return message %s: \n", method);
	    }
	    else {
		smtpd_respond_multiline(scr, method_ret, in->msgs);
	    }
	}
	else {
	    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			 "[mod_smtpd_auth.c] smtpd_auth_unrecognized_command: No handler for %s: \n", method);
	}

	apr_pool_destroy(pool);
    }

    return ret;
}

smtpd_retcode smtpd_auth_pre_connect(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_auth_config *conf = ap_get_module_config(scr->s->module_config, &smtpd_auth_module);
    
    if (conf != NULL && conf->enable == 0) {
	return SMTPD_DECLINED;
    }

    conf->methods = apr_hash_make(scr->p);

    return SMTPD_DECLINED;
}

smtpd_retcode smtpd_auth_connect(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_auth_config *conf = ap_get_module_config(scr->s->module_config, &smtpd_auth_module);
    apr_pool_t *pool;
    apr_hash_index_t *hi;
    const char *str = "AUTH";

    if (conf == NULL || conf->enable < 1 || apr_hash_count(conf->methods) == 0) {
	return SMTPD_DECLINED;
    }

    apr_pool_create(&pool, scr->p);

    for (hi = apr_hash_first(pool, conf->methods);
	 hi;
	 hi = apr_hash_next(hi)) {
	const void *key;
	void *val;

	apr_hash_this(hi, &key, NULL, &val);

	str = apr_pstrcat(pool, str, " ", key, NULL);
    }

    smtpd_register_extension(scr, str);

    apr_pool_destroy(pool);

    return SMTPD_DECLINED;
}

APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd_auth, SMTPD_AUTH, int, password,
				    (smtpd_conn_rec *scr, smtpd_auth_lookup_rec *lookup_rec),
				    (scr, lookup_rec),
				    0,0);

void smtpd_auth_register_method(smtpd_conn_rec *scr, char *key, smtpd_auth_method *func)
{
    smtpd_auth_config *conf = ap_get_module_config(scr->s->module_config, &smtpd_auth_module);

    if (conf != NULL ) {
	char *dupkey = apr_pstrdup(scr->p, key);
	smtpd_auth_method_st *hand = apr_palloc(scr->p, sizeof(*hand));

	hand->func = func;

	str_toupper(dupkey);
	apr_hash_set(conf->methods, dupkey, APR_HASH_KEY_STRING, hand);
    }
}

static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, connect, smtpd_auth_pre_connect, NULL, NULL,
		      APR_HOOK_REALLY_FIRST);

    APR_OPTIONAL_HOOK(smtpd, connect, smtpd_auth_connect, NULL, NULL,
		      APR_HOOK_MIDDLE);
    
    APR_OPTIONAL_HOOK(smtpd, unrecognized_command, smtpd_auth_unrecognized_command,
		      NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA smtpd_auth_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_auth_config_new,   /* server config */
    smtpd_auth_config_merge, /* merge server config */
    smtpd_auth_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
