/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"
#include "apr_base64.h"
#include "apr_md5.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include "mod_smtpd.h"
#include "mod_smtpd_auth.h"

module AP_MODULE_DECLARE_DATA smtpd_authm_crammd5_module;

static void hex_str(char *dst, char *src, int len)
{
    char hex[] = "0123456789abcdef";
    int  i;
    for (i = 0; i < len; i++)
    {
	dst[2 * i] = hex[(src[i] & 0xf0) >> 4];
	dst[2 * i + 1] = hex[src[i] & 0x0f];
    }
}

typedef struct {
    int enable;
} smtpd_authm_crammd5_config;

static void *smtpd_authm_crammd5_config_new(apr_pool_t *p,
					  server_rec *s)
{
    smtpd_authm_crammd5_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_crammd5_config));

    cfg->enable = -1;

    return cfg;
}

static void *smtpd_authm_crammd5_config_merge(apr_pool_t *p, void *b,
					      void *n)
{
    smtpd_authm_crammd5_config *base = (smtpd_authm_crammd5_config *)b;
    smtpd_authm_crammd5_config *new = (smtpd_authm_crammd5_config *)n;

    smtpd_authm_crammd5_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_crammd5_config));

    if (new->enable == -1) {
	cfg->enable = base->enable;
    }
    else {
	cfg->enable = new->enable;
    }

    return cfg;
}

static const char *smtpd_authm_crammd5_enable(cmd_parms *cmd,
					    void *in_dconf, int flag)
{
    smtpd_authm_crammd5_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_authm_crammd5_module);

    conf->enable = flag;

    return NULL;
}

static const command_rec smtpd_auth_cmds[] =
{
    AP_INIT_FLAG("SmtpAuthMethodCrammd5Enable", smtpd_authm_crammd5_enable,
		 (void *) NULL, RSRC_CONF,
		 "Enable the CRAM-MD5 auth method"),
    {NULL}
};

SMTP_AUTH_METHOD_DECLARE(crammd5)
{
    smtpd_retcode ret = SMTPD_DECLINED;
    apr_pool_t *pool;
    char **newstr;

    apr_size_t  buf_size = 1024;
    apr_size_t  buf_len = 0;
    char       *buf;

    char       *challenge;
    apr_size_t  challenge_len;
    char       *challenge_encoded ;

    char *nonce;
    int   nonce_hex_len;
    char *nonce_hex;

    apr_pool_create(&pool, scr->p);
    buf  = apr_palloc(pool, buf_size + 1);

    nonce = apr_pcalloc(pool, 33);
    nonce_hex_len = apr_base64_encode_len(32);
    nonce_hex = apr_pcalloc(pool, nonce_hex_len);

    apr_generate_random_bytes(nonce, 32);
    nonce[32] = '\0';
    
    apr_base64_encode(nonce_hex, nonce, 32);

    challenge = apr_psprintf(pool, "<%s@%s>", nonce_hex,
			     scr->c->remote_host ? scr->c->remote_host : scr->c->remote_ip );

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_authm_crammd5: challenge: %s", challenge);

    challenge_len = strlen(challenge);
    challenge_encoded  = apr_pcalloc(pool, apr_base64_encode_len(challenge_len));

    apr_base64_encode(challenge_encoded, challenge, challenge_len);

    smtpd_respond_oneline(scr, 334, challenge_encoded);

    if (smtpd_getline(scr, buf, buf_size, &buf_len) == APR_SUCCESS) {
	char       *user;
	apr_size_t  user_len;
	apr_size_t  user_size;

	char       *response;
	apr_size_t  response_len;

	smtpd_auth_lookup_rec *lookup_rec;

	user_size = apr_base64_decode_len(buf);
	user = apr_pcalloc(pool, user_size + 1);
	user_len = apr_base64_decode(user, buf);

	response = strchr(user, ' ');

	response[0] = '\0';
	response++;
	response_len = strlen(response);

	lookup_rec = apr_pcalloc(pool, sizeof(smtpd_auth_lookup_rec) );

	lookup_rec->pool = pool;
	lookup_rec->user = user;
	lookup_rec->user_len = strlen(user);

	if (smtpd_auth_run_password(scr, lookup_rec) == 1) {
	    char k_ipad[65];    /* inner padding - key XORd with ipad */
	    char k_opad[65];    /* outer padding - key XORd with opad */
	    char tk[APR_MD5_DIGESTSIZE];
	    char digest[APR_MD5_DIGESTSIZE];
	    char hex_digest[33];
	    apr_md5_ctx_t context;
	    int i;

	    const char *key = lookup_rec->pass;
	    int   key_len = lookup_rec->pass_len;

	    bzero(hex_digest, sizeof(hex_digest));

	    if ( lookup_rec->pass_len > 64 )
	    {
		apr_md5(tk, lookup_rec->pass, lookup_rec->pass_len);

		key = tk;
		key_len = APR_MD5_DIGESTSIZE;
	    }

	    /*
	     * the HMAC_MD5 transform looks like:
	     *
	     * MD5(K XOR opad, MD5(K XOR ipad, text))
	     *
	     * where K is an n byte key
	     * ipad is the byte 0x36 repeated 64 times
	     * opad is the byte 0x5c repeated 64 times
	     * and text is the data being protected
	     */

	    /* start out by storing key in pads */
	    bzero( k_ipad, sizeof k_ipad );
	    bzero( k_opad, sizeof k_opad );
	    bcopy( key, k_ipad, key_len );
	    bcopy( key, k_opad, key_len );

	    /* XOR key with ipad and opad values */
	    for (i=0; i < 64; i++)
	    {
		k_ipad[i] ^= 0x36;
		k_opad[i] ^= 0x5c;
	    }

	    /* perform inner MD5 */
	    apr_md5_init( &context );                   /* init context for 1st + pass */
	    apr_md5_update( &context, k_ipad, 64 );     /* start with inner pad */
	    apr_md5_update( &context, challenge, challenge_len ); /* then text of datagram */
	    apr_md5_final( digest, &context );          /* finish up 1st pass */

	    /* perform outer MD5 */
	    apr_md5_init( &context );                   /* init context for 2nd +  pass */
	    apr_md5_update( &context, k_opad, 64 );     /* start with outer pad */
	    apr_md5_update( &context, digest, APR_MD5_DIGESTSIZE );     /* then results of 1st + hash */
	    apr_md5_final( digest, &context );          /* finish up 2nd pass */

	    hex_str(hex_digest, digest, APR_MD5_DIGESTSIZE);

	    if (strcmp(hex_digest, response) == 0) {
		in->msgs = apr_array_make(in->p, 1, sizeof(char*));
		newstr = (char**)apr_array_push(in->msgs);
		*newstr = apr_pstrdup(in->p, "OK Authenticated");
		ret = 235;
/*
		scr->transaction->auth = apr_table_make(scr->transaction->p, 2);
		apr_table_set(scr->transaction->auth, "user", user);
		apr_table_set(scr->transaction->auth, "method", "CRAM-MD5");
*/

		ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
			     "smtpd_auth: auth cram-md5 authenticated: %s", user);
	    }
	    else {
		in->msgs = apr_array_make(in->p, 1, sizeof(char*));
		newstr = (char**)apr_array_push(in->msgs);
		*newstr = apr_pstrdup(in->p, "Authentication failed");
		ret = 535;

		ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			     "smtpd_auth: auth cram-md5 authentication failed (%s)! %s : %s", user, hex_digest, response);
	    }

	}
	else {
	    in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	    newstr = (char**)apr_array_push(in->msgs);
	    *newstr = apr_pstrdup(in->p, "Authentication failed");
	    ret = 535;

	    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			 "smtpd_auth: auth cram-md5 authentication failed (%s)!", user);
	}
    }
    else {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrdup(in->p, "Missing input");
	ret = 501;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_auth: auth cram-md5 missing input!");
    }

    apr_pool_destroy(pool);

    return ret;
}

smtpd_retcode smtpd_auth_crammd5_connect(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_authm_crammd5_config *conf =
	ap_get_module_config(scr->s->module_config, &smtpd_authm_crammd5_module);

    if ( conf && conf->enable == 1 ) {
	smtpd_auth_register_method(scr, "CRAM-MD5", SMTP_AUTH_METHOD_FUNC(crammd5));
    }

    return SMTPD_DECLINED;
}


static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, connect, smtpd_auth_crammd5_connect, NULL, NULL,
		      APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA smtpd_authm_crammd5_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_authm_crammd5_config_new,   /* server config */
    smtpd_authm_crammd5_config_merge, /* merge server config */
    smtpd_auth_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
