/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"
#include "apr_base64.h"
#include "apr_md5.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include "mod_smtpd.h"
#include "mod_smtpd_auth.h"

module AP_MODULE_DECLARE_DATA smtpd_authm_digestmd5_module;

static void hex_str(char *dst, char *src, int len)
{
    char hex[] = "0123456789abcdef";
    int  i;
    for (i = 0; i < len; i++)
    {
	dst[2 * i] = hex[(src[i] & 0xf0) >> 4];
	dst[2 * i + 1] = hex[src[i] & 0x0f];
    }
}

typedef struct {
    int enable;
} smtpd_authm_digestmd5_config;

static void *smtpd_authm_digestmd5_config_new(apr_pool_t *p,
					  server_rec *s)
{
    smtpd_authm_digestmd5_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_digestmd5_config));

    cfg->enable = -1;

    return cfg;
}

static void *smtpd_authm_digestmd5_config_merge(apr_pool_t *p, void *b,
				     void *n)
{
    smtpd_authm_digestmd5_config *base = (smtpd_authm_digestmd5_config *)b;
    smtpd_authm_digestmd5_config *new = (smtpd_authm_digestmd5_config *)n;

    smtpd_authm_digestmd5_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_digestmd5_config));

    if (new->enable == -1) {
	cfg->enable = base->enable;
    }
    else {
	cfg->enable = new->enable;
    }

    return cfg;
}

static const char *smtpd_authm_digestmd5_enable(cmd_parms *cmd,
					    void *in_dconf, int flag)
{
    smtpd_authm_digestmd5_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_authm_digestmd5_module);

    conf->enable = flag;

    return NULL;
}

static const command_rec smtpd_auth_cmds[] =
{
    AP_INIT_FLAG("SmtpAuthMethodDigestmd5Enable", smtpd_authm_digestmd5_enable,
		 (void *) NULL, RSRC_CONF,
		 "Enable the DIGEST-MD5 auth method"),
    {NULL}
};

SMTP_AUTH_METHOD_DECLARE(digestmd5)
{
    smtpd_retcode ret = SMTPD_DECLINED;
    apr_pool_t *pool;
    char **newstr;

    apr_size_t  buf_size = 1024;
    apr_size_t  buf_len = 0;
    char       *buf;

    const char *prompt;
    apr_size_t  prompt_len;
    char       *prompt_encoded;

    char *nonce;
    int   nonce_hex_len;
    char *nonce_hex;

    apr_pool_create(&pool, scr->p);

    buf = apr_palloc(pool, buf_size + 1);

    nonce = apr_pcalloc(pool, 65);
    nonce_hex_len = apr_base64_encode_len(64);
    nonce_hex = apr_pcalloc(pool, nonce_hex_len+1);

    apr_generate_random_bytes(nonce, 64);
    nonce[64] = '\0';

    apr_base64_encode(nonce_hex, nonce, 64);

    prompt = apr_psprintf(pool, "realm=\"%s\",nonce=\"%s\",qop=\"auth\",algorithm=md5-sess",
			  scr->c->remote_host ? scr->c->remote_host : scr->c->remote_ip, nonce_hex);
    prompt_len = strlen(prompt);

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_auth: auth digest-md5: prompt: %s", prompt);

    prompt_encoded = apr_pcalloc(pool, apr_base64_encode_len(prompt_len));

    apr_base64_encode(prompt_encoded, prompt, prompt_len);

    smtpd_respond_oneline(scr, 334, prompt_encoded);

    if (smtpd_getline(scr, buf, buf_size, &buf_len) == APR_SUCCESS) {
	char       *challenge;
	apr_size_t  challenge_len;
	apr_size_t  challenge_size;
	apr_hash_t *hash = apr_hash_make( pool );

	char *key;
	char *tok_state;
	char *str;

	smtpd_auth_lookup_rec *lookup_rec;

	challenge_size = apr_base64_decode_len(buf);
	challenge = apr_pcalloc(pool, challenge_size + 1);
	challenge_len = apr_base64_decode(challenge, buf);

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_auth: auth digest-md5: challenge: %s", challenge);

	str = apr_pstrdup(pool, challenge);

	for (key = apr_strtok(str, ",", &tok_state );
	     key;
	     key = apr_strtok(NULL, ",", &tok_state))
	{
	    char *val = strchr(key, '=');

	    if (val != NULL)
	    {
		char *k = apr_pstrndup(pool, key, val-key);
		char *v = apr_pstrdup(pool, val + 1);
		int   v_len = strlen(v);

		if (v[0] == '"') {
		    v++;
		    v_len--;
		}

		if (v[v_len-1] == '"') {
		    v[v_len-1] = '\0';
		}

		ap_str_tolower(k);
		apr_hash_set(hash, k, val-key, v );
	    }
	}

	lookup_rec = apr_pcalloc(pool, sizeof(smtpd_auth_lookup_rec) );

	lookup_rec->pool = pool;
	lookup_rec->user = apr_pstrcat(pool, apr_hash_get(hash, "username", APR_HASH_KEY_STRING),
				       "@", apr_hash_get(hash, "realm", APR_HASH_KEY_STRING), NULL );
	lookup_rec->user_len = strlen(lookup_rec->user);

	if (smtpd_auth_run_password(scr, lookup_rec) == 1) {
	    char *user = apr_hash_get(hash, "username", APR_HASH_KEY_STRING);
	    char *realm = apr_hash_get(hash, "realm", APR_HASH_KEY_STRING);
	    char hv[APR_MD5_DIGESTSIZE];
	    char hv_hex[33];
	    char *nonce = apr_hash_get(hash, "nonce", APR_HASH_KEY_STRING);
	    int   nonce_len = strlen(nonce);
	    char *cnonce = apr_hash_get(hash, "cnonce", APR_HASH_KEY_STRING);
	    int   cnonce_len = strlen(cnonce);
	    char *url = apr_hash_get(hash, "digest-uri", APR_HASH_KEY_STRING);
	    char *qop = apr_hash_get(hash, "qop", APR_HASH_KEY_STRING);
	    char *nc = apr_hash_get(hash, "nc", APR_HASH_KEY_STRING);
	    char a1[APR_MD5_DIGESTSIZE];
	    char a1_hex[33];
	    char a2[APR_MD5_DIGESTSIZE];
	    char a2_hex[33];
	    char resp[APR_MD5_DIGESTSIZE];
	    char resp_hex[33];

	    char *response = apr_hash_get(hash, "response", APR_HASH_KEY_STRING);

	    char *tmp;
	    int   tmp_len;

	    bzero(hv_hex, sizeof(hv_hex));
	    bzero(a1_hex, sizeof(a1_hex));
	    bzero(a2_hex, sizeof(a2_hex));
	    bzero(resp_hex, sizeof(resp_hex));

	    tmp = apr_pstrcat(pool, user, ":", realm, ":", lookup_rec->pass, NULL);
	    tmp_len = strlen(tmp);

	    apr_md5(hv, tmp, tmp_len);

	    tmp_len = APR_MD5_DIGESTSIZE + 1 + nonce_len + 1 + cnonce_len;
	    tmp = apr_pcalloc(pool, tmp_len + 1 );

	    memcpy(tmp, hv, APR_MD5_DIGESTSIZE);
	    tmp[APR_MD5_DIGESTSIZE] = ':';
	    memcpy(tmp+APR_MD5_DIGESTSIZE+1, nonce, nonce_len);
	    tmp[APR_MD5_DIGESTSIZE+1+nonce_len] = ':';
	    memcpy(tmp+APR_MD5_DIGESTSIZE+1+nonce_len+1, cnonce, cnonce_len);

	    apr_md5(a1, tmp, tmp_len);

	    hex_str(a1_hex, a1, APR_MD5_DIGESTSIZE);

	    tmp = apr_pstrcat(pool, "AUTHENTICATE:", url, NULL);
	    tmp_len = strlen(tmp);

	    apr_md5(a2, tmp, tmp_len);

	    hex_str(a2_hex, a2, APR_MD5_DIGESTSIZE);

	    tmp = apr_pstrcat(pool, a1_hex, ":", nonce, ":", nc, ":", cnonce, ":", qop, ":", a2_hex, NULL);
	    tmp_len = strlen(tmp);

	    apr_md5(resp, tmp, tmp_len);

	    hex_str(resp_hex, resp, APR_MD5_DIGESTSIZE);

	    if ( strcmp(resp_hex, response) == 0 ) {
		in->msgs = apr_array_make(in->p, 1, sizeof(char*));
		newstr = (char**)apr_array_push(in->msgs);
		*newstr = apr_pstrdup(in->p, "OK Authenticated");
		ret = 235;
/*
		scr->transaction->auth = apr_table_make(scr->transaction->p, 2);
		apr_table_set(scr->transaction->auth, "user", user);
		apr_table_set(scr->transaction->auth, "method", "DIGEST-MD5");
*/

		ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
			     "smtpd_auth: auth digest-md5 authenticated: %s@%s", user, realm );
	    }
	    else {
		in->msgs = apr_array_make(in->p, 1, sizeof(char*));
		newstr = (char**)apr_array_push(in->msgs);
		*newstr = apr_pstrdup(in->p, "Authentication failed");
		ret = 535;

		ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			     "smtpd_auth: auth digest-md5 authentication failed: %s@%s : %s : %s", user, realm, resp_hex, response);
	    }

	}
	else {
	    in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	    newstr = (char**)apr_array_push(in->msgs);
	    *newstr = apr_pstrdup(in->p, "Authentication failed");
	    ret = 535;

	    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			 "smtpd_auth: auth digest-md5 authentication failed!");
	}
    }
    else {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrdup(in->p, "Missing input");
	ret = 501;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_auth: auth digest-md5 missing input!");
    }

    apr_pool_destroy(pool);

    return ret;
}

smtpd_retcode smtpd_auth_digestmd5_connect(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_authm_digestmd5_config *conf =
	ap_get_module_config(scr->s->module_config, &smtpd_authm_digestmd5_module);

    if ( conf && conf->enable == 1 ) {
	smtpd_auth_register_method(scr, "DIGEST-MD5", SMTP_AUTH_METHOD_FUNC(digestmd5));
    }

    return SMTPD_DECLINED;
}


static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, connect, smtpd_auth_digestmd5_connect, NULL, NULL,
		      APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA smtpd_authm_digestmd5_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_authm_digestmd5_config_new,   /* server config */
    smtpd_authm_digestmd5_config_merge, /* merge server config */
    smtpd_auth_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
