/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"
#include "apr_base64.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include "mod_smtpd.h"
#include "mod_smtpd_auth.h"

module AP_MODULE_DECLARE_DATA smtpd_authm_login_module;

typedef struct {
    int enable;
} smtpd_authm_login_config;

static void *smtpd_authm_login_config_new(apr_pool_t *p,
					  server_rec *s)
{
    smtpd_authm_login_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_login_config));

    cfg->enable = -1;

    return cfg;
}

static void *smtpd_authm_login_config_merge(apr_pool_t *p, void *b,
				     void *n)
{
    smtpd_authm_login_config *base = (smtpd_authm_login_config *)b;
    smtpd_authm_login_config *new = (smtpd_authm_login_config *)n;

    smtpd_authm_login_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_login_config));

    if (new->enable == -1) {
	cfg->enable = base->enable;
    }
    else {
	cfg->enable = new->enable;
    }

    return cfg;
}

static const char *smtpd_authm_login_enable(cmd_parms *cmd,
					    void *in_dconf, int flag)
{
    smtpd_authm_login_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_authm_login_module);

    conf->enable = flag;

    return NULL;
}

static const command_rec smtpd_auth_cmds[] =
{
    AP_INIT_FLAG("SmtpAuthMethodLoginEnable", smtpd_authm_login_enable,
		 (void *) NULL, RSRC_CONF,
		 "Enable the LOGIN auth method"),
    {NULL}
};

SMTP_AUTH_METHOD_DECLARE(login)
{
    smtpd_retcode ret = SMTPD_DECLINED;
    apr_pool_t *pool;
    char **newstr;

    char       *user;
    apr_size_t  user_len;

    apr_size_t  buf_size = 1024;
    apr_size_t  buf_len = 0;
    char       *buf;

    char  *command = data;

    const char *prompt;
    apr_size_t  prompt_len;
    apr_size_t  prompt_encoded_size;
    char       *prompt_encoded;

    apr_pool_create(&pool, scr->p);
    buf  = apr_palloc(pool, buf_size + 1);

    if (command == NULL || command[0] == '\0') {
	prompt = "username: ";
	prompt_len = strlen(prompt);
	prompt_encoded_size = apr_base64_encode_len(prompt_len);
	prompt_encoded  = apr_pcalloc(pool, prompt_encoded_size + 1);

	apr_base64_encode(prompt_encoded, prompt, prompt_len);

	smtpd_respond_oneline(scr, 334, prompt_encoded);

	if (smtpd_getline(scr, buf, buf_size, &buf_len) == APR_SUCCESS) {
	    user = apr_pcalloc(pool, apr_base64_decode_len(buf));
	    user_len = apr_base64_decode(user, buf);
	}
	else {
	    in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	    newstr = (char**)apr_array_push(in->msgs);
	    *newstr = apr_pstrdup(in->p, "Missing input");
	    ret = 501;

	    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
			 "smtpd_auth: auth login missing input!");

	    goto auth_exit;
	}
    }
    else {
	user = apr_pcalloc(pool, apr_base64_decode_len(command));
	user_len = apr_base64_decode(user, command);
    }

    prompt = "password: ";
    prompt_len = strlen(prompt);

    prompt_encoded_size = apr_base64_encode_len(prompt_len);
    prompt_encoded  = apr_pcalloc(pool, prompt_encoded_size);

    apr_base64_encode(prompt_encoded, prompt, strlen(prompt));

    smtpd_respond_oneline(scr, 334, prompt_encoded);

    if (smtpd_getline(scr, buf, buf_size, &buf_len) == APR_SUCCESS) {
	char       *pass;
	apr_size_t  pass_len;
	apr_size_t  pass_size;
	smtpd_auth_lookup_rec *lookup_rec;

	pass_size = apr_base64_decode_len(buf);
	pass = apr_pcalloc(pool, pass_size + 1);
	pass_len = apr_base64_decode(pass, buf);

	lookup_rec = apr_pcalloc(pool, sizeof(smtpd_auth_lookup_rec) );

	lookup_rec->pool = pool;
	lookup_rec->user = user;
	lookup_rec->pass = pass;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_auth: auth login got: user '%s' pass '%s'", user, pass);

	if (smtpd_auth_run_password(scr, lookup_rec) == 1) {
	    in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	    newstr = (char**)apr_array_push(in->msgs);
	    *newstr = apr_pstrdup(in->p, "OK Authenticated");
	    ret = 235;
/*
	    scr->transaction->auth = apr_table_make(scr->transaction->p, 3);
	    apr_table_set(scr->transaction->auth, "user", user);
	    apr_table_set(scr->transaction->auth, "pass", pass);
	    apr_table_set(scr->transaction->auth, "method", "LOGIN");
*/
	    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
			 "smtpd_auth: auth login authenticated: %s", lookup_rec->user);
	}
	else {
	    in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	    newstr = (char**)apr_array_push(in->msgs);
	    *newstr = apr_pstrdup(in->p, "Authentication failed");
	    ret = 535;

	    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
			 "smtpd_auth: auth login authentication failed!");
	}
    }
    else {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrdup(in->p, "Missing input");
	ret = 501;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_auth: auth login missing input!");
    }

 auth_exit:

    apr_pool_destroy(pool);

    return ret;
}

smtpd_retcode smtpd_authm_login_connect(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_authm_login_config *conf =
	ap_get_module_config(scr->s->module_config, &smtpd_authm_login_module);

    if ( conf && conf->enable == 1 ) {
	smtpd_auth_register_method(scr, "LOGIN", SMTP_AUTH_METHOD_FUNC(login));
    }

    return SMTPD_DECLINED;
}


static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, connect, smtpd_authm_login_connect, NULL, NULL,
		      APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA smtpd_authm_login_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_authm_login_config_new,   /* server config */
    smtpd_authm_login_config_merge, /* merge server config */
    smtpd_auth_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
