/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"
#include "apr_base64.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include "mod_smtpd.h"
#include "mod_smtpd_auth.h"

module AP_MODULE_DECLARE_DATA smtpd_authm_plain_module;

typedef struct {
    int enable;
} smtpd_authm_plain_config;

static void *smtpd_authm_plain_config_new(apr_pool_t *p,
					  server_rec *s)
{
    smtpd_authm_plain_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_plain_config));

    cfg->enable = -1;

    return cfg;
}

static void *smtpd_authm_plain_config_merge(apr_pool_t *p, void *b,
				     void *n)
{
    smtpd_authm_plain_config *base = (smtpd_authm_plain_config *)b;
    smtpd_authm_plain_config *new = (smtpd_authm_plain_config *)n;

    smtpd_authm_plain_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_authm_plain_config));

    if (new->enable == -1) {
	cfg->enable = base->enable;
    }
    else {
	cfg->enable = new->enable;
    }

    return cfg;
}

static const char *smtpd_authm_plain_enable(cmd_parms *cmd,
					    void *in_dconf, int flag)
{
    smtpd_authm_plain_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_authm_plain_module);

    conf->enable = flag;

    return NULL;
}

static const command_rec smtpd_authm_plain_cmds[] =
{
    AP_INIT_FLAG("SmtpAuthMethodPlainEnable", smtpd_authm_plain_enable,
		 (void *) NULL, RSRC_CONF,
		 "Enable the PLAIN auth method"),
    {NULL}
};

SMTP_AUTH_METHOD_DECLARE(plain)
{
    smtpd_retcode ret = SMTPD_DECLINED;
    const char *user;
    const char *pass;

    char *decode;
    char *decode_end;
    apr_size_t decode_len;

    char **newstr;

    smtpd_auth_lookup_rec *lookup_rec;

    apr_pool_t *pool;

    if (data == NULL || *(char *)data == '\0') {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrdup(in->p, "Missing input");
	ret = 501;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_authm_plain: missing input!");
	return ret;
    }

    apr_pool_create(&pool, scr->p);

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_auth: auth data: %s", (char *)data);

    decode = apr_pcalloc(pool, apr_base64_decode_len(data));

    decode_len = apr_base64_decode(decode, data);

    user = decode;
    decode_end = decode + decode_len;

    while ( *user != '\0' && user < decode_end)
    {
	user++;
    }

    if ( user + 1 >= decode_end ) {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrcat(in->p, "Can not decode AUTH parameter ", (char *)data, NULL);
	ret = 501;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_authm_plain: Can not decode AUTH parameter (user) %s", (char *)data);
	return ret;
    }

    user++;

    pass = user;
    while (*pass != '\0' && pass < decode_end) {
	pass++;
    }

    if ( pass + 1 >= decode_end ) {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrcat(in->p, "Can not decode AUTH parameter ", (char *)data, NULL);
	ret = 501;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_authm_plain: Can not decode AUTH parameter (pass) %s", (char *)data);
	return ret;
    }

    pass++;

    lookup_rec = apr_pcalloc(pool, sizeof(smtpd_auth_lookup_rec) );

    lookup_rec->pool = pool;
    lookup_rec->user = user;
    lookup_rec->user_len = strlen(user);
    lookup_rec->pass = pass;
    lookup_rec->pass_len = strlen(pass);

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_auth: auth plain got: user '%s' pass '%s'", user, pass);

    if (smtpd_auth_run_password(scr, lookup_rec) == 1) {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrdup(in->p, "OK Authenticated");
	ret = 235;

/*
	scr->transaction->auth = apr_table_make(scr->transaction->p, 3);
	apr_table_set(scr->transaction->auth, "user", user);
	apr_table_set(scr->transaction->auth, "pass", pass);
	apr_table_set(scr->transaction->auth, "method", "PLAIN");
*/

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_auth: auth plain authenticated: %s", lookup_rec->user);
    }
    else {
	in->msgs = apr_array_make(in->p, 1, sizeof(char*));
	newstr = (char**)apr_array_push(in->msgs);
	*newstr = apr_pstrdup(in->p, "Authentication failed");
	ret = 535;

	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_auth: auth plain authentication failed!");
    }

    apr_pool_destroy(pool);

    return ret;
}

smtpd_retcode smtpd_authm_plain_connect(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_authm_plain_config *conf =
	ap_get_module_config(scr->s->module_config, &smtpd_authm_plain_module);

    if ( conf && conf->enable == 1 ) {
	smtpd_auth_register_method(scr, "PLAIN", SMTP_AUTH_METHOD_FUNC(plain));
    }

    return SMTPD_DECLINED;
}


static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, connect, smtpd_authm_plain_connect, NULL, NULL,
		      APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA smtpd_authm_plain_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_authm_plain_config_new,   /* server config */
    smtpd_authm_plain_config_merge, /* merge server config */
    smtpd_authm_plain_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
