/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"
#include "apr_support.h"
#include "apr_md5.h"            /* for apr_password_validate */
#include "apr_lib.h"            /* for apr_isspace */
#include "apr_base64.h"         /* for apr_base64_decode et al */
#define APR_WANT_STRFUNC        /* for strcasecmp */
#include "apr_want.h"
#include "apr_dbd.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_provider.h"

#include "mod_smtpd.h"
#include "mod_dbd.h"
#include "mod_smtpd_auth.h"

module AP_MODULE_DECLARE_DATA smtpd_authv_dbd_module;

static void *smtpd_authv_dbd_config_new(apr_pool_t *p, server_rec *s)
{
    return NULL;
}

static void *smtpd_authv_dbd_config_merge(apr_pool_t *p,
					 void *b, void *n)
{
    if (n != NULL) {
	return apr_pstrdup(p, n);
    }

    if (b != NULL) {
	return apr_pstrdup(p, b);
    }

    return NULL;
}

static const char *smtpd_authv_dbd_cfg(cmd_parms *cmd, void *dummy,
					 const char *arg)
{
    ap_set_module_config(cmd->server->module_config,
			 &smtpd_authv_dbd_module,
			 apr_pstrdup(cmd->pool, arg) );

    return NULL;
}

static const command_rec smtpd_authv_dbd_cmds[] =
{
    AP_INIT_TAKE1("SmtpAuthDbdEnable", smtpd_authv_dbd_cfg,
		  (void *) NULL, RSRC_CONF,
		  "The name of the dbd to use for queries"),
    {NULL}
};

int smtpd_authv_dbd(smtpd_conn_rec *scr, smtpd_auth_lookup_rec *lookup_rec)
{
    ap_dbd_t *dbd;
    apr_dbd_prepared_t *statement;
    apr_dbd_results_t *res = NULL;
    apr_dbd_row_t *row = NULL;

    apr_pool_t *pool;

    const char *val = NULL;
    int ret = 0;

    const char *label = ap_get_module_config(scr->s->module_config,
					     &smtpd_authv_dbd_module);

    if ( label == NULL ) {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_authv_dbd]: no dbd prepare statement setup! %s",
		     lookup_rec->user);
	return 0;
    }

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_authv_dbd: user: %s",  lookup_rec->user);

    apr_pool_create(&pool, lookup_rec->pool);

    dbd = ap_dbd_open(pool, scr->s);

    if (dbd == NULL) {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_authv_dbd]: DBD connection was not acquired! %s",
		     lookup_rec->user);
	goto authv_dbd_exit;
    }

    statement = apr_hash_get(dbd->prepared, label, APR_HASH_KEY_STRING);
    if (statement == NULL) {
	ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, scr->s,

		     label, lookup_rec->user);
	goto authv_dbd_exit;
    }

    if (apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, statement,
			 0, lookup_rec->user, NULL) != 0) {
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_authv_dbd: Error looking up %s in database", lookup_rec->user);
	goto authv_dbd_exit;
    }

    if (apr_dbd_get_row(dbd->driver, pool, res, &row, -1) != 0) {
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_authv_dbd: nothing found for: %s", lookup_rec->user );
	goto authv_dbd_exit;
    }

    if ((val = apr_dbd_get_entry(dbd->driver, row, 0)) == NULL) {
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		     "smtpd_authv_dbd: found NULL for %s", lookup_rec->user );
	goto authv_dbd_exit;
    }

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_authv_dbd: found %s for %s", val, lookup_rec->user);

    if ( lookup_rec->pass == NULL ) {
	lookup_rec->pass = apr_pstrdup(lookup_rec->pool, val);
	lookup_rec->pass_len = strlen(lookup_rec->pass);
	ret = 1;
    }
    else if (strcmp(lookup_rec->pass, val) == 0) {
	ret = 1;
    }

authv_dbd_exit:

    apr_pool_destroy(pool);

    return ret;
}

static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd_auth, password, smtpd_authv_dbd,
		      NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA smtpd_authv_dbd_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_authv_dbd_config_new,   /* server config */
    smtpd_authv_dbd_config_merge, /* merge server config */
    smtpd_authv_dbd_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
